defmodule Mpx.Http do
  @moduledoc """
  Provides the base http functionality for Ministry PLatform access including GET, POST, PUT and DELETE
  """
  import Mpx.Config

  @endpoint Application.get_env(:mpx, :mp_base_url, "")

  @typedoc """
  Specifies what is returned from all of the HTTP calls
  """
  @type t :: {atom(), map() | list()} | {atom(), term()}


  @doc """
  Wrapper for the Ministry Platform GET Api call.

  Takes optional keyword list of options:
    `[ "$select": "The_columns_to_select",
       "$filter": "The filter to apply",
       etc...
    ]`
  See swagger for a complete list of parameters that the MP rest api takes.

  Returns an `HTTPoison.Response` or `HTTPoison.Error`
  """
  @spec get(binary(), binary(), keyword(binary)) :: t
  def get(path, auth_token, opts \\ []) do
    "#{from_config(:mp_base_url, @endpoint)}/ministryplatformapi/#{path}#{stringify_url_options(opts)}"
    |> HTTPoison.get(%{"Authorization" => "Bearer #{auth_token}"})
    |> map_response
  end

  @doc """
  Makes calls to the the `ministryplatformapi/oauth` endpoint. This can be used for getting the current logged in user by passing `me` and the current token.
  """
  @spec oauth_get(binary(), binary()) :: t
  def oauth_get(path, auth_token) do
    IO.puts "verifying auth token"
    "#{from_config(:mp_base_url, @endpoint)}/ministryplatformapi/oauth/connect/#{path}"
    |> HTTPoison.get(%{"Authorization" => "Bearer #{auth_token}"})
    |> map_response
  end

  @doc """
  Wrapper for the Ministry Platform POST Api call

  Takes optional keyword list of options:
    `[ "$select": "The_columns_to_select",
       "$filter": "The filter to apply",
       etc...
    ]`
  see swagger for a complete list of parameters that the MP rest api takes.

  Returns an `HTTPoison.Response` or `HTTPoison.Error`
  """
  @spec post(binary(), binary(), map() | list(), keyword(binary())) :: t
  def post(path, auth_token, data, opts \\ []) do
    case Poison.encode(data) do
      {:ok, encoded_data} ->
        "#{from_config(:mp_base_url, @endpoint)}/ministryplatformapi/#{path}#{stringify_url_options(opts)}"
        |> HTTPoison.post(encoded_data,
                          %{"Authorization" => "Bearer #{auth_token}"})
        |> map_response
      err -> err
    end
  end

  @doc """
  Wrapper for the Ministry Platform PUT Api call

  Takes optional keyword list of options:
    `[ "$select": "The_columns_to_select",
       "$filter": "The filter to apply",
       etc...
    ]`
  see the swagger documentation for a
  complete list of parameters that the MP rest api takes.

  Returns an `HTTPoison.Response` or `HTTPoison.Error`
  """
  @spec put(binary(), binary(), map() | list(), keyword(binary())) :: t
  def put(path, auth_token, data, opts \\ []) do
    case Poison.encode(data) do
      {:ok, encoded_data} ->
        "#{from_config(:mp_base_url, @endpoint)}/ministryplatformapi/#{path}#{stringify_url_options(opts)}"
        |> HTTPoison.put(encoded_data, %{"Authorization" => "Bearer #{auth_token}"})
        |> map_response
      err -> err
    end
  end

  @doc """
  Wrapper for the Ministry Platform DELETE Api call

  Takes optional keyword list of options:
    `[ "$select": "The_columns_to_select",
       "$filter": "The filter to apply",
       etc...
    ]`
  see swagger for a complete list of parameters that the MP rest api takes

  Returns an `HTTPoison.Response` or `HTTPoison.Error`
  """
  @spec delete(binary(), binary(), Mpx.Tables.delete_options) :: t
  def delete(path, auth_token, opts \\ []) do
    "#{from_config(:mp_base_url, @endpoint)}/ministryplatformapi/#{path}#{stringify_url_options(opts)}"
    |> HTTPoison.delete(%{"Authorization" => "Bearer #{auth_token}"})
    |> map_response
  end

  @doc """
  Attempts to get a token from MP based on the username/password/client_id/client_secret passed in

  If an empty list is passed as the authentication options, attempts to use the configuration file
  """
  @spec authenticate_user(Mpx.Authentication.authentication_opts | []) :: {:ok | :error, String.t}
  def authenticate_user(opts \\ []) do
    "#{from_config(:mp_base_url, @endpoint)}/ministryplatformapi/oauth/connect/token"
     |> HTTPoison.post(
       Mpx.Authentication.Body.get_auth_body(opts),
       %{"Content-type" => "application/x-www-form-urlencoded"})
     |> map_auth_response
  end

  @doc """
  Attempts to authenticate a client against MP
  """
  @spec authenticate_client([client_id: String.t, client_secret: String.t]) :: {:ok | :error, String.t}
  def authenticate_client(opts) do
    "#{from_config(:mp_base_url, @endpoint)}/ministryplatformapi/oauth/connect/token"
    |> HTTPoison.post(
      Mpx.Authentication.Body.get_auth_body(opts),
      %{"Content-type" => "application/x-www-form-urlencoded"})
    |> map_auth_response
  end

  @doc false
  @spec stringify_url_options(keyword() | []) :: binary()
  defp stringify_url_options([]), do: ""
  defp stringify_url_options(opts) do
    options =
      opts
      |> unique_keys
      |> Enum.reduce([], fn(key, acc) ->
        vals =
          opts
          |> Keyword.get_values(key)
          |> Enum.map(fn(x) -> URI.encode("#{key}=#{x}") end)
            acc ++ vals
      end)
      |> Enum.join("&")
    "?#{options}"
  end

  @doc false
  @spec map_response({:ok, HTTPoison.Response.t} | {:error, HTTPoison.Error}) :: {:ok, map() | list()} | {:error, binary()}
  defp map_response({:ok, %HTTPoison.Response{body: body, status_code: 200}}), do: Poison.decode(body)
  defp map_response({:ok, %HTTPoison.Response{body: body, status_code: 401}}), do: {:error, :unauthorized}
  defp map_response({:ok, %HTTPoison.Response{body: body, status_code: other}}) do
    res =
      body
      |> Poison.decode
      |> translate_error
    {:error, res}
  end
  defp map_response({:error, %HTTPoison.Error{id: id, reason: reason}}), do: {:error, reason}

  @doc false
  defp translate_error({:ok, decoded}), do: decoded
  defp translate_error({:error, reason}), do: :unknown


  @doc false
  defp map_auth_response({:ok, %HTTPoison.Response{status_code: 200, body: body}}) do
    with {:ok, token} <- Poison.decode(body) do
      {:ok, Map.get(token, "access_token")}
    else
      {:error, {:invalid, line, char}} ->
        {:error, "unable to decode token from " <> body <> " because of " <> char <> " at line " <> line}
      {:error, err} ->
        {:error, "unable to decode token from " <> body <> " because of " <> err}
    end
  end
  defp map_auth_response({:ok, %HTTPoison.Response{status_code: code, body: body}}), do: {:error, "unable to authenticate"}
  defp map_auth_response({:error, %HTTPoison.Error{reason: reason}}), do: {:error, reason}

  @spec unique_keys(Keyword.t) :: list()
  defp unique_keys(keyword_list) do
    keyword_list
    |> Enum.uniq_by(fn {k, _} -> k end)
    |> Keyword.keys
  end
end
